##############################################################################
#
# Copyright (c) 2007 Zope Foundation and Contributors.
# All Rights Reserved.
#
# This software is subject to the provisions of the Zope Public License,
# Version 2.1 (ZPL).  A copy of the ZPL should accompany this distribution.
# THIS SOFTWARE IS PROVIDED "AS IS" AND ANY AND ALL EXPRESS OR IMPLIED
# WARRANTIES ARE DISCLAIMED, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF TITLE, MERCHANTABILITY, AGAINST INFRINGEMENT, AND FITNESS
# FOR A PARTICULAR PURPOSE.
#
##############################################################################
"""Page Drawing Related Element Processing
"""
import types
import reportlab.pdfbase.pdfform
from z3c.rml import attr, directive, interfaces, occurence

try:
    import reportlab.graphics.barcode
except ImportError:
    # barcode package has not been installed
    import types
    import reportlab.graphics
    reportlab.graphics.barcode = types.ModuleType('barcode')
    reportlab.graphics.barcode.getCodeNames = lambda : ()


class IBarCodeBase(interfaces.IRMLDirectiveSignature):
    """Create a bar code."""

    code = attr.Choice(
        title=u'Code',
        description=u'The name of the type of code to use.',
        choices=reportlab.graphics.barcode.getCodeNames(),
        required=True)

    value = attr.TextNode(
        title=u'Value',
        description=u'The value represented by the code.',
        required=True)

    width = attr.Measurement(
        title=u'Width',
        description=u'The width of the barcode.',
        required=False)

    height = attr.Measurement(
        title=u'Height',
        description=u'The height of the barcode.',
        required=False)

    barStrokeColor = attr.Color(
        title=u'Bar Stroke Color',
        description=(u'The color of the line strokes in the barcode.'),
        required=False)

    barStrokeWidth = attr.Measurement(
        title=u'Bar Stroke Width',
        description=u'The width of the line strokes in the barcode.',
        required=False)

    barFillColor = attr.Color(
        title=u'Bar Fill Color',
        description=(u'The color of the filled shapes in the barcode.'),
        required=False)

    gap = attr.Measurement(
        title=u'Gap',
        description=u'The width of the inter-character gaps.',
        required=False)

    # Bar code dependent attributes
    # I2of5, Code128, Standard93, FIM, POSTNET, Ean13B
    barWidth = attr.Measurement(
        title=u'Bar Width',
        description=u'The width of the smallest bar within the barcode',
        required=False)

    # I2of5, Code128, Standard93, FIM, POSTNET
    barHeight = attr.Measurement(
        title=u'Bar Height',
        description=u'The height of the symbol.',
        required=False)

    # I2of5
    ratio = attr.Float(
        title=u'Ratio',
        description=(u'The ratio of wide elements to narrow elements. '
                     u'Must be between 2.0 and 3.0 (or 2.2 and 3.0 if the '
                     u'barWidth is greater than 20 mils (.02 inch)).'),
        min=2.0,
        max=3.0,
        required=False)

    # I2of5
    # Should be boolean, but some code want it as int; will still work
    checksum = attr.Integer(
        title=u'Ratio',
        description=(u'A flag that enables the computation and inclusion of '
                     u'the check digit.'),
        required=False)

    # I2of5
    bearers = attr.Float(
        title=u'Bearers',
        description=(u'Height of bearer bars (horizontal bars along the top '
                     u'and bottom of the barcode). Default is 3 '
                     u'x-dimensions. Set to zero for no bearer bars.'
                     u'(Bearer bars help detect misscans, so it is '
                     u'suggested to leave them on).'),
        required=False)

    # I2of5, Code128, Standard93, FIM, Ean13
    quiet = attr.Boolean(
        title=u'Quiet Zone',
        description=(u'A flag to include quiet zones in the symbol.'),
        required=False)

    # I2of5, Code128, Standard93, FIM, Ean13
    lquiet = attr.Measurement(
        title=u'Left Quiet Zone',
        description=(u"Quiet zone size to the left of code, if quiet is "
                     u"true. Default is the greater of .25 inch or .15 times "
                     u"the symbol's length."),
        required=False)

    # I2of5, Code128, Standard93, FIM, Ean13
    rquiet = attr.Measurement(
        title=u'Right Quiet Zone',
        description=(u"Quiet zone size to the right of code, if quiet is "
                     u"true. Default is the greater of .25 inch or .15 times "
                     u"the symbol's length."),
        required=False)

    # I2of5, Code128, Standard93, FIM, POSTNET, Ean13
    fontName = attr.String(
        title=u'Font Name',
        description=(u'The font used to print the value.'),
        required=False)

    # I2of5, Code128, Standard93, FIM, POSTNET, Ean13
    fontSize = attr.Measurement(
        title=u'Font Size',
        description=(u'The size of the value text.'),
        required=False)

    # I2of5, Code128, Standard93, FIM, POSTNET, Ean13
    humanReadable = attr.Boolean(
        title=u'Human Readable',
        description=(u'A flag when set causes the value to be printed below '
                     u'the bar code.'),
        required=False)

    # I2of5, Standard93
    stop = attr.Boolean(
        title=u'Show Start/Stop',
        description=(u'A flag to specify whether the start/stop symbols '
                     u'are to be shown.'),
        required=False)

    # FIM, POSTNET
    spaceWidth = attr.Measurement(
        title=u'Space Width',
        description=u'The space of the inter-character gaps.',
        required=False)

    # POSTNET
    shortHeight = attr.Measurement(
        title=u'Short Height',
        description=u'The height of the short bar.',
        required=False)

    # Ean13
    textColor = attr.Color(
        title=u'Text Color',
        description=(u'The color of human readable text.'),
        required=False)

    # USPS4S
    routing = attr.String(
        title=u'Routing',
        description=u'The routing information string.',
        required=False)

    # QR
    barLevel = attr.Choice(
        title=u'Bar Level',
        description=u'The error correction level for QR code',
        choices=['L', 'M', 'Q', 'H'],
        required=False)

    barBorder = attr.Measurement(
        title=u'Bar Border',
        description=u'The width of the border around a QR code.',
        required=False)


class IBarCode(IBarCodeBase):
    """A barcode graphic."""

    x = attr.Measurement(
        title=u'X-Position',
        description=u'The x-position of the lower-left corner of the barcode.',
        default=0,
        required=False)

    y = attr.Measurement(
        title=u'Y-Position',
        description=u'The y-position of the lower-left corner of the barcode.',
        default=0,
        required=False)

    isoScale = attr.Boolean(
        title=u'Isometric Scaling',
        description=u'When set, the aspect ration of the barcode is enforced.',
        required=False)

class BarCode(directive.RMLDirective):
    signature = IBarCode

    def process(self):
        kw = dict(self.getAttributeValues())
        name = kw.pop('code')
        kw['value'] = str(kw['value'])
        x = kw.pop('x', 0)
        y = kw.pop('y', 0)
        code = reportlab.graphics.barcode.createBarcodeDrawing(name, **kw)
        manager = attr.getManager(self, interfaces.ICanvasManager)
        code.drawOn(manager.canvas, x, y)


class IField(interfaces.IRMLDirectiveSignature):
    """A field."""

    title = attr.Text(
        title=u'Title',
        description=u'The title of the field.',
        required=True)

    x = attr.Measurement(
        title=u'X-Position',
        description=u'The x-position of the lower-left corner of the field.',
        default=0,
        required=True)

    y = attr.Measurement(
        title=u'Y-Position',
        description=u'The y-position of the lower-left corner of the field.',
        default=0,
        required=True)


class Field(directive.RMLDirective):
    signature = IField
    callable = None
    attrMapping = {}

    def process(self):
        kwargs = dict(self.getAttributeValues(attrMapping=self.attrMapping))
        canvas = attr.getManager(self, interfaces.ICanvasManager).canvas
        getattr(reportlab.pdfbase.pdfform, self.callable)(canvas, **kwargs)


class ITextField(IField):
    """A text field within the PDF"""

    width = attr.Measurement(
        title=u'Width',
        description=u'The width of the text field.',
        required=True)

    height = attr.Measurement(
        title=u'Height',
        description=u'The height of the text field.',
        required=True)

    value = attr.Text(
        title=u'Value',
        description=u'The default text value of the field.',
        required=False)

    maxLength = attr.Integer(
        title=u'Maximum Length',
        description=u'The maximum amount of characters allowed in the field.',
        required=False)

    multiline = attr.Boolean(
        title=u'Multiline',
        description=u'A flag when set allows multiple lines within the field.',
        required=False)

class TextField(Field):
    signature = ITextField
    callable = 'textFieldAbsolute'
    attrMapping = {'maxLength': 'maxlen'}


class IButtonField(IField):
    """A button field within the PDF"""

    value = attr.Choice(
        title=u'Value',
        description=u'The value of the button.',
        choices=('Yes', 'Off'),
        required=True)

class ButtonField(Field):
    signature = IButtonField
    callable = 'buttonFieldAbsolute'


class IOption(interfaces.IRMLDirectiveSignature):
    """An option in the select field."""

    value = attr.TextNode(
        title=u'Value',
        description=u'The value of the option.',
        required=True)

class Option(directive.RMLDirective):
    signature = IOption

    def process(self):
        value = self.getAttributeValues(valuesOnly=True)[0]
        self.parent.options.append(value)


class ISelectField(IField):
    """A selection field within the PDF"""
    occurence.containing(
        occurence.ZeroOrMore('option', IOption))

    width = attr.Measurement(
        title=u'Width',
        description=u'The width of the select field.',
        required=True)

    height = attr.Measurement(
        title=u'Height',
        description=u'The height of the select field.',
        required=True)

    value = attr.Text(
        title=u'Value',
        description=u'The default value of the field.',
        required=False)

class SelectField(Field):
    signature = ISelectField
    callable = 'selectFieldAbsolute'
    factories = {'option': Option}

    def process(self):
        self.options = []
        self.processSubDirectives()
        kwargs = dict(self.getAttributeValues(attrMapping=self.attrMapping))
        kwargs['options'] = self.options
        canvas = attr.getManager(self, interfaces.ICanvasManager).canvas
        getattr(reportlab.pdfbase.pdfform, self.callable)(canvas, **kwargs)



##############################################################################
#
# Copyright (c) 2007 Zope Foundation and Contributors.
# All Rights Reserved.
#
# This software is subject to the provisions of the Zope Public License,
# Version 2.1 (ZPL).  A copy of the ZPL should accompany this distribution.
# THIS SOFTWARE IS PROVIDED "AS IS" AND ANY AND ALL EXPRESS OR IMPLIED
# WARRANTIES ARE DISCLAIMED, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF TITLE, MERCHANTABILITY, AGAINST INFRINGEMENT, AND FITNESS
# FOR A PARTICULAR PURPOSE.
#
##############################################################################
"""RML Directive Implementation
"""
import logging
import zope.interface
import zope.schema

from lxml import etree
from z3c.rml import interfaces
from z3c.rml.attr import getManager

logging.raiseExceptions = False
logger = logging.getLogger("z3c.rml")

ABORT_ON_INVALID_DIRECTIVE = False

def DeprecatedDirective(iface, reason):
    zope.interface.directlyProvides(iface, interfaces.IDeprecatedDirective)
    iface.setTaggedValue('deprecatedReason', reason)
    return iface

def getFileInfo(directive, element=None):
    root = directive
    while root.parent:
        root = root.parent
    if element is None:
        element = directive.element
    return '(file %s, line %i)' %(root.filename, element.sourceline)

@zope.interface.implementer(interfaces.IRMLDirective)
class RMLDirective(object):
    signature = None
    factories = {}

    def __init__(self, element, parent):
        self.element = element
        self.parent = parent

    def getAttributeValues(self, ignore=None, select=None, attrMapping=None,
                           includeMissing=False, valuesOnly=False):
        """See interfaces.IRMLDirective"""
        manager = getManager(self)
        cache = '%s.%s' % (self.signature.__module__, self.signature.__name__)
        if cache in manager.attributesCache:
            fields = manager.attributesCache[cache]
        else:
            fields = []
            for name, attr in zope.schema.getFieldsInOrder(self.signature):
                fields.append((name, attr))
            manager.attributesCache[cache] = fields

        items = []
        for name, attr in fields:
            # Only add the attribute to the list, if it is supposed there
            if ((ignore is None or name not in ignore) and
                (select is None or name in select)):
                # Get the value.
                value = attr.bind(self).get()
                # If no value was found for a required field, raise a value
                # error
                if attr.required and value is attr.missing_value:
                    raise ValueError(
                        'No value for required attribute "%s" '
                        'in directive "%s" %s.' % (
                        name, self.element.tag, getFileInfo(self)))
                # Only add the entry if the value is not the missing value or
                # missing values are requested to be included.
                if value is not attr.missing_value or includeMissing:
                    items.append((name, value))

        # Sort the items based on the section
        if select is not None:
            select = list(select)
            items = sorted(items, key=lambda n: select.index(n[0]))

        # If the attribute name does not match the internal API
        # name, then convert the name to the internal one
        if attrMapping:
            items = [(attrMapping.get(name, name), value)
                     for name, value in items]

        # Sometimes we only want the values without the names
        if valuesOnly:
            return [value for name, value in items]

        return items

    def processSubDirectives(self, select=None, ignore=None):
        # Go through all children of the directive and try to process them.
        for element in self.element.getchildren():
            # Ignore all comments
            if isinstance(element, etree._Comment):
                continue
            # Raise an error/log any unknown directive.
            if element.tag not in self.factories:
                msg = "Directive %r could not be processed and was " \
                      "ignored. %s" %(element.tag, getFileInfo(self, element))
                # Record any tags/elements that could not be processed.
                logger.warning(msg)
                if ABORT_ON_INVALID_DIRECTIVE:
                    raise ValueError(msg)
                continue
            if select is not None and element.tag not in select:
                continue
            if ignore is not None and element.tag in ignore:
                continue
            directive = self.factories[element.tag](element, self)
            directive.process()

    def process(self):
        self.processSubDirectives()



##############################################################################
#
# Copyright (c) 2007 Zope Foundation and Contributors.
# All Rights Reserved.
#
# This software is subject to the provisions of the Zope Public License,
# Version 2.1 (ZPL).  A copy of the ZPL should accompany this distribution.
# THIS SOFTWARE IS PROVIDED "AS IS" AND ANY AND ALL EXPRESS OR IMPLIED
# WARRANTIES ARE DISCLAIMED, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF TITLE, MERCHANTABILITY, AGAINST INFRINGEMENT, AND FITNESS
# FOR A PARTICULAR PURPOSE.
#
##############################################################################
"""RML to PDF Converter
"""
import os
import six
import sys
import zope.interface
from lxml import etree
from z3c.rml import document, interfaces
import argparse

zope.interface.moduleProvides(interfaces.IRML2PDF)


def parseString(xml, removeEncodingLine=True, filename=None):
    if isinstance(xml, six.text_type) and removeEncodingLine:
        # RML is a unicode string, but oftentimes documents declare their
        # encoding using <?xml ...>. Unfortuantely, I cannot tell lxml to
        # ignore that directive. Thus we remove it.
        if xml.startswith('<?xml'):
            xml = xml.split('\n', 1)[-1]
    root = etree.fromstring(xml)
    doc = document.Document(root)
    if filename:
        doc.filename = filename
    output = six.BytesIO()
    doc.process(output)
    output.seek(0)
    return output


def go(xmlInputName, outputFileName=None, outDir=None, dtdDir=None):
    if dtdDir is not None:
        sys.stderr.write('The ``dtdDir`` option is not yet supported.\n')

    if hasattr(xmlInputName, 'read'):
        # it is already a file-like object
        xmlFile = xmlInputName
        xmlInputName = 'input.pdf'
    else:
        xmlFile = open(xmlInputName, 'rb')
    root = etree.parse(xmlFile).getroot()
    doc = document.Document(root)
    doc.filename = xmlInputName

    outputFile = None

    # If an output filename is specified, create an output filepointer for it
    if outputFileName is not None:
        if hasattr(outputFileName, 'write'):
            # it is already a file-like object
            outputFile = outputFileName
            outputFileName = 'output.pdf'
        else:
            if outDir is not None:
                outputFileName = os.path.join(outDir, outputFileName)
            outputFile = open(outputFileName, 'wb')

    # Create a Reportlab canvas by processing the document
    doc.process(outputFile)

    if outputFile:
        outputFile.close()
    xmlFile.close()


def main(args=None):
    if args is None:
        parser = argparse.ArgumentParser(
            prog='rml2pdf',
            description='Converts file in RML format into PDF file.',
            epilog='Copyright (c) 2007 Zope Foundation and Contributors.'
        )
        parser.add_argument('xmlInputName', help='RML file to be processed')
        parser.add_argument('outputFileName', nargs='?', help='output PDF file name')
        parser.add_argument('outDir', nargs='?', help='output directory')
        parser.add_argument('dtdDir', nargs='?', help='directory with XML DTD (not yet supported)')
        pargs = parser.parse_args()
        args = (pargs.xmlInputName, pargs.outputFileName, pargs.outDir, pargs.dtdDir)

    go(*args)


if __name__ == '__main__':
    canvas = go(sys.argv[1])



##############################################################################
#
# Copyright (c) 2007 Zope Foundation and Contributors.
# All Rights Reserved.
#
# This software is subject to the provisions of the Zope Public License,
# Version 2.1 (ZPL).  A copy of the ZPL should accompany this distribution.
# THIS SOFTWARE IS PROVIDED "AS IS" AND ANY AND ALL EXPRESS OR IMPLIED
# WARRANTIES ARE DISCLAIMED, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF TITLE, MERCHANTABILITY, AGAINST INFRINGEMENT, AND FITNESS
# FOR A PARTICULAR PURPOSE.
#
##############################################################################
"""Page Drawing Related Element Processing
"""
import six
from z3c.rml import attr, directive, interfaces

try:
    import PyPDF2
    from PyPDF2.generic import NameObject
except ImportError:
    # We don't want to require pyPdf, if you do not want to use the features
    # in this module.
    PyPDF2 = None


class MergePostProcessor(object):

    def __init__(self):
        self.operations = {}

    def process(self, inputFile1):
        input1 = PyPDF2.PdfFileReader(inputFile1)
        output = PyPDF2.PdfFileWriter()
        # TODO: Do not access protected classes
        output._info.getObject().update(input1.documentInfo)

        # Add the outlines, if any
        if "/Outlines" in input1.trailer["/Root"]:
            if output._root:
                # Backwards-compatible with PyPDF2 version 1.21
                output._root.getObject()[NameObject("/Outlines")] = (
                    output._addObject(input1.trailer["/Root"]["/Outlines"]))
            else:
                # Compatible with PyPDF2 version 1.22+
                output._root_object[NameObject("/Outlines")] = (
                    output._addObject(input1.trailer["/Root"]["/Outlines"]))

        for (num, page) in enumerate(input1.pages):
            if num in self.operations:
                for mergeFile, mergeNumber in self.operations[num]:
                    merger = PyPDF2.PdfFileReader(mergeFile)
                    mergerPage = merger.getPage(mergeNumber)
                    mergerPage.mergePage(page)
                    page = mergerPage
            output.addPage(page)

        outputFile = six.BytesIO()
        output.write(outputFile)
        return outputFile


class IMergePage(interfaces.IRMLDirectiveSignature):
    """Merges an existing PDF Page into the one to be generated."""

    filename = attr.File(
        title=u'File',
        description=(u'Reference to the PDF file to extract the page from.'),
        required=True)

    page = attr.Integer(
        title=u'Page Number',
        description=u'The page number of the PDF file that is used to merge..',
        required=True)


class MergePage(directive.RMLDirective):
    signature = IMergePage

    def getProcessor(self):
        manager = attr.getManager(self, interfaces.IPostProcessorManager)
        procs = dict(manager.postProcessors)
        if 'MERGE' not in procs:
            proc = MergePostProcessor()
            manager.postProcessors.append(('MERGE', proc))
            return proc
        return procs['MERGE']

    def process(self):
        if PyPDF2 is None:
            raise Exception(
                'pyPdf is not installed, so this feature is not available.')
        inputFile, inPage = self.getAttributeValues(valuesOnly=True)
        manager = attr.getManager(self, interfaces.ICanvasManager)
        outPage = manager.canvas.getPageNumber()-1

        proc = self.getProcessor()
        pageOperations = proc.operations.setdefault(outPage, [])
        pageOperations.append((inputFile, inPage))


class MergePageInPageTemplate(MergePage):

    def process(self):
        if PyPDF2 is None:
            raise Exception(
                'pyPdf is not installed, so this feature is not available.')
        inputFile, inPage = self.getAttributeValues(valuesOnly=True)

        onPage = self.parent.pt.onPage
        def drawOnCanvas(canvas, doc):
            onPage(canvas, doc)
            outPage = canvas.getPageNumber()-1
            proc = self.getProcessor()
            pageOperations = proc.operations.setdefault(outPage, [])
            pageOperations.append((inputFile, inPage))

        self.parent.pt.onPage = drawOnCanvas



##############################################################################
#
# Copyright (c) 2007 Zope Foundation and Contributors.
# All Rights Reserved.
#
# This software is subject to the provisions of the Zope Public License,
# Version 2.1 (ZPL).  A copy of the ZPL should accompany this distribution.
# THIS SOFTWARE IS PROVIDED "AS IS" AND ANY AND ALL EXPRESS OR IMPLIED
# WARRANTIES ARE DISCLAIMED, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF TITLE, MERCHANTABILITY, AGAINST INFRINGEMENT, AND FITNESS
# FOR A PARTICULAR PURPOSE.
#
##############################################################################
"""Condition Implementation
"""
import reportlab
import six
import sys
import zope.interface
import zope.schema
from zope.schema import fieldproperty

class ICondition(zope.interface.Interface):
    """Condition that is checked before a  directive is available."""

    __doc__ = zope.schema.TextLine(
        title=u'Description',
        description=u'The description of the condition.',
        required=True)

    def __call__(directive):
        """Check whether the condition is fulfilled for the given directive."""


class IOccurence(zope.interface.Interface):
    """Description of the occurence of a sub-directive."""

    __doc__ = zope.schema.TextLine(
        title=u'Description',
        description=u'The description of the occurence.',
        required=True)

    if six.PY2:
        tag = zope.schema.BytesLine(
            title=u'Tag',
            description=u'The tag of the sub-directive within the directive',
            required=True)
    else:
        tag = zope.schema.TextLine(
            title=u'Tag',
            description=u'The tag of the sub-directive within the directive',
            required=True)

    signature = zope.schema.Field(
        title=u'Signature',
        description=u'The signature of the sub-directive.',
        required=True)

    condition = zope.schema.Field(
        title=u'Condition',
        description=u'The condition that the directive is available.',
        required=False)


@zope.interface.implementer(ICondition)
def laterThanReportlab21(directive):
    """The directive is only available in Reportlab 2.1 and higher."""
    return [int(num) for num in reportlab.Version.split('.')] >= (2, 0)


def containing(*occurences):
    frame = sys._getframe(1)
    f_locals = frame.f_locals
    f_globals = frame.f_globals

    if not (f_locals is not f_globals
            and f_locals.get('__module__')
            and f_locals.get('__module__') == f_globals.get('__name__')
            ):
        raise TypeError("contains not called from signature interface")

    f_locals['__interface_tagged_values__'] = {'directives': occurences}


@zope.interface.implementer(IOccurence)
class Occurence(object):
    tag = fieldproperty.FieldProperty(IOccurence['tag'])
    signature = fieldproperty.FieldProperty(IOccurence['signature'])
    condition = fieldproperty.FieldProperty(IOccurence['condition'])

    def __init__(self, tag, signature, condition=None):
        self.tag = tag
        self.signature = signature
        self.condition = condition


class ZeroOrMore(Occurence):
    """Zero or More

    This sub-directive can occur zero or more times.
    """

class ZeroOrOne(Occurence):
    """Zero or one

    This sub-directive can occur zero or one time.
    """

class OneOrMore(Occurence):
    """One or More

    This sub-directive can occur one or more times.
    """

class One(Occurence):
    """One

    This sub-directive must occur exactly one time.
    """
